##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HSDS (HDF5 Scalable Data Service), Libraries and      #
# Utilities.  The full HSDS copyright notice, including                      #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import sys
import logging
    
try:
    import h5py 
    import h5pyd 
except ImportError as e:
    sys.stderr.write("ERROR : %s : install it to use this utility...\n" % str(e)) 
    sys.exit(1)
 
if __name__ == "utillib":
    from chunkiter import ChunkIterator
else:
    from .chunkiter import ChunkIterator
  
def is_h5py(objref):
    # Return True if objref is a h5py reference and False is not 
    if objref and isinstance(objref.id.id, int):
        return True
    else:
        return False

#----------------------------------------------------------------------------------
def copy_attribute(desobj, name, srcobj, verbose=False):
    msg = "creating attribute {} in {}".format(name, srcobj.name)
    logging.debug(msg)
    if verbose:
        print(msg)
    try:
        # TBD: we are potentially losing some fidelity here by accessing 
        # the attribute as a numpy array and then passing that to the 
        # the create method.  Better would be to use the h5py low-level
        # API to ensure we get the exact type of the attribute.
        desobj.attrs.create(name, srcobj.attrs[name])
    except (IOError, TypeError) as e:
        msg = "ERROR: failed to create attribute {} of object {} -- {}".format(name, desobj.name, str(e))
        logging.error(msg)
        print(msg)
# copy_attribute
      
#----------------------------------------------------------------------------------
def create_dataset(fd, dobj, verbose=False, nodata=False):
    """ create a dataset using the properties of the passed in h5py dataset.
        If successful, proceed to copy attributes and data.
    """
    msg = "creating dataset {}, shape: {}, type: {}".format(dobj.name, dobj.shape, dobj.dtype)
    logging.info(msg)
    if verbose:
        print(msg) 
       
    fillvalue = None
    try:    
        # can trigger a runtime error if fillvalue is undefined
        fillvalue = dobj.fillvalue
    except RuntimeError:
        pass  # ignore
    chunks=None
    if dobj.chunks:
        chunks = tuple(dobj.chunks)
    try:
        dset = fd.create_dataset( dobj.name, shape=dobj.shape, dtype=dobj.dtype, chunks=chunks, \
                compression=dobj.compression, shuffle=dobj.shuffle, \
                fletcher32=dobj.fletcher32, maxshape=dobj.maxshape, \
                compression_opts=dobj.compression_opts, fillvalue=fillvalue, \
                scaleoffset=dobj.scaleoffset)
        msg = "dataset created, uuid: {}, chunk_size: {}".format(dset.id.id, str(dset.chunks))  
        logging.info(msg)
        if verbose:
            print(msg)
    except (IOError, TypeError) as e:
        msg = "ERROR: failed to create dataset: {}".format(str(e))
        logging.error(msg)
        print(msg)
        return
    # create attributes
    for da in dobj.attrs:
        copy_attribute(dset, da, dobj)

    if nodata:
        msg = "skipping data load"
        logging.info(msg)
        if verbose:
            print(msg)
        return

    msg = "iterating over chunks for {}".format(dobj.name)
    logging.info(msg)
    if verbose:
        print(msg)
    try:
        it = ChunkIterator(dset)

        logging.debug("src dtype: {}".format(dobj.dtype))
        logging.debug("des dtype: {}".format(dset.dtype))
        
        for s in it:
            msg = "writing dataset data for slice: {}".format(s)
            logging.info(msg)
            if verbose:
                print(msg)
            arr = dobj[s]
            dset[s] = arr
    except (IOError, TypeError) as e:
        msg = "ERROR : failed to copy dataset data : {}".format(str(e))
        logging.error(msg)
        print(msg)
    msg = "done with dataload for {}".format(dobj.name)
    logging.info(msg)
    if verbose:
        print(msg)
# create_dataset

#----------------------------------------------------------------------------------
def create_group(fd, gobj, verbose=False):
    msg = "creating group {}".format(gobj.name)
    logging.info(msg)
    if verbose:
        print(msg)
    grp = fd.create_group(gobj.name)

    # create attributes
    for ga in gobj.attrs:
        copy_attribute(grp, ga, gobj)

    # create any soft/external links
    for title in gobj:
        lnk = gobj.get(title, getlink=True)
        link_classname = lnk.__class__.__name__
        if link_classname == "HardLink":
            logging.debug("Got hardlink: {}".format(title))
            # TBD: handle the case where multiple hardlinks point to same object
        elif link_classname == "SoftLink":
            msg = "creating SoftLink({}) with title: {}".format(lnk.path, title)
            if verbose:
                print(msg)
            logging.info(msg)
            if is_h5py(fd):
                soft_link = h5py.SoftLink(lnk.path)
            else:
                soft_link = h5pyd.SoftLink(lnk.path)
            grp[title] = soft_link
        elif link_classname == "ExternalLink":
            msg = "creating ExternalLink({}, {}) with title: {}".format(lnk.filename, lnk.path, title)
            if verbose:
                print(msg)
            logging.info(msg)
            if is_h5py(fd):
                ext_link = h5py.ExternalLink(lnk.filename, lnk.path)
            else:
                ext_link = h5pyd.ExternalLink(lnk.filename, lnk.path)
            grp[title] = ext_link
        else:
            msg = "Unexpected link type: {}".format(lnk.__class__.__name__)
            logging.warning(msg)
            if verbose:
                print(msg)
# create_group

#----------------------------------------------------------------------------------
def create_datatype(fd, obj, verbose=False):
    msg = "creating datatype {}".format(obj.name)
    logging.info(msg)
    if verbose:
        print(msg)
    fd[obj.name] = obj.dtype
    ctype = fd[obj.name]
    # create attributes
    for ga in obj.attrs:
        copy_attribute(ctype, ga, obj)
# create_datatype
      
#----------------------------------------------------------------------------------
def load_file(fin, fout, verbose=False, nodata=False):
    logging.info("input file: {}".format(fin.filename))   
    logging.info("output file: {}".format(fout.filename))
     
    # create any root attributes
    for ga in fin.attrs:
        copy_attribute(fout, ga, fin)

    def object_create_helper(name, obj):
        class_name = obj.__class__.__name__
        if class_name == "Dataset":
            create_dataset(fout, obj, verbose=verbose, nodata=nodata)
        elif class_name == "Group":
            create_group(fout, obj, verbose=verbose)
        elif class_name == "Datatype":
            create_datatype(fout, obj, verbose=verbose)
        else:
            logging.error("no handler for object class: {}".format(type(obj)))

    # build a rough map of the file using the internal function above
    fin.visititems(object_create_helper)
        
    # Fully flush the h5py handle.  
    fout.close() 
      
    # close up the source domain, see reason(s) for this below
    fin.close() 
    msg="load_file complete: {}"
    logging.info(msg)
    if verbose:
        print(msg)
    
    return 0
# load_file
    
  




##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

import numpy as np
import six
import config

if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase


class TestAttribute(TestCase):


    def test_create(self):
        filename = self.getFileName("create_attribute")
        print("filename:", filename)
        f = h5py.File(filename, 'w')

        #f.attrs['a1'] = 42  #  to-do fix

        g1 = f.create_group('g1')

        g1.attrs['a1'] = 42
         
        n = g1.attrs['a1']
        self.assertEqual(n, 42)

        self.assertEqual(len(g1.attrs), 1)

        g1.attrs['b1'] = list(range(10))

        # try replacing 'a1'
        g1.attrs['a1'] = 24

        self.assertEqual(len(g1.attrs), 2)
        
        # create an attribute with explict UTF type
        dt = h5py.special_dtype(vlen=str)
        g1.attrs.create('c1', "Hello HDF", dtype=dt)
         

        value = g1.attrs['c1']

        self.assertEqual(value, "Hello HDF")

        # create attribute with implicit UTF type
        g1.attrs.create('d1', "This is a python string")

        attr_names = []
        for a in g1.attrs:
            attr_names.append(a)
        self.assertEqual(len(attr_names), 4)
        self.assertTrue('a1' in attr_names)
        self.assertTrue('b1' in attr_names)
        self.assertTrue('c1' in attr_names)
        self.assertTrue('d1' in attr_names)

        # create a array attribute
        g1.attrs["ones"] = np.ones((10,))
        arr = g1.attrs["ones"]
        self.assertEqual(arr.shape, (10,))
        for i in range(10):
            self.assertEqual(arr[i], 1)

        # array of strings
        g1.attrs['strings'] = ["Hello", "Good-bye"]
        arr = g1.attrs['strings']
        self.assertEqual(arr.shape, (2,))
        self.assertEqual(arr[0], "Hello")
        self.assertEqual(arr[1], "Good-bye")
        if six.PY3:
            self.assertEqual(arr.dtype, h5py.special_dtype(vlen=str))
        else:
            self.assertEqual(arr.dtype, np.dtype("S8"))

        # byte values
        g1.attrs['e1'] = b"Hello"

        # close file
        f.close()




         

if __name__ == '__main__':
    ut.main()







##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of H5Serv (HDF5 REST Server) Service, Libraries and      #
# Utilities.  The full HDF5 REST Server copyright notice, including          #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
import json
import config
if config.get("use_h5py"):
    import h5py
else:
    import h5pyd as h5py

from common import ut, TestCase


class TestObjRef(TestCase):


    def test_create(self):
        filename = self.getFileName("objref_test")
        print(filename)
        f = h5py.File(filename, 'w')
        self.assertTrue(f.id.id is not None)
        self.assertTrue('/' in f)
        r = f['/']

        r.create_group('g1')
        self.assertTrue('g1' in r)
        g1 = r['g1']


        g11 = g1.create_group('g1.1')

        g11_ref = g11.ref
        #print("g11_ref:", g11_ref)
        #print("uuid:", g11_ref.id.uuid)
        #print("domain:", g11_ref.id.domain)
        #print("type:", g11_ref.id.objtype_code)
        #print("g11_ref_tolist:", g11_ref.tolist())

        # todo - fix
        #self.assertTrue(isinstance(g11_ref, h5py.Reference))


        r.create_group('g2')
        self.assertEqual(len(r), 2)
        g2 = r['g2']
        #print("json dump:", json.dumps(g2.__dict__))
        """
        g11ref = g2[g11_ref]
        #print("g11ref:", g11ref)
        #print("g11ref name:", g11ref.name)
        #print("g11ref type:", type(g11ref))
        g11ref.create_group("foo")
        """
        d1 = g2.create_dataset('d1', (10,), dtype='i8')

        d1_ref = d1.ref

        dt = h5py.special_dtype(ref=h5py.Reference)
        #print("dt:", dt)
        #print("dt.kind:", dt.kind)
        #print("dt.meta:", dt.metadata['ref'])
        print("special_dtype:", type(dt.metadata['ref']))
        self.assertTrue(dt.metadata['ref'] is h5py.Reference)

        dset = g1.create_dataset('myrefs', (10,), dtype=dt)
        #print("dset kind:", dset.dtype.kind)
        #print("dset.dtype.kind:", dset.dtype.kind)
        ref = h5py.check_dtype(ref=dset.dtype)
        #print("check_dtype:", ref)
        null_ref = dset[0]
        #print("null_ref:", null_ref)
        dset[0] = g11_ref
        dset[1] = d1_ref
        #g2.attrs['dataset'] = dset.ref

        # todo - references as data will need h5pyd equivalent of h5t module
        # g2.attrs.create('dataset', dset.ref, dtype=dt)
        #print("g11_ref type:", type(g11_ref))
        a_ref = dset[0]
        #print("a_ref", type(a_ref) )

        f.close()

if __name__ == '__main__':
    ut.main()




